package com.sillycycle.bagleyd.cubes;

/*
 * @(#)CubesSolve.java
 *
 * Taken from Puzzle originally written in C
 * $XConsortium: puzzle.c,v 1.14 94/03/28 18:34:10 gildea Exp $
 * by Don Bennett, HP Labs
 *
 * Copyright 1987, 1988  Don Bennett
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * Copyright 2003 - 2021  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import javax.swing.JFrame;

/**
 * The <code>CubesSolve</code> class is the code that does the work
 * of solving the puzzle.  (Commonly seen as a 4x4 grid of sliding
 * piece numbered 1-15 with one empty space.)  Needs work on being
 * able to solve puzzles less than a 4x4 and for 3D.
 *
 * The idea for the solution algorithm - solving the puzzle
 * in layers working from the outside in - comes indirectly from
 * John Nagle.
 */

class CubesSolve {
	private static JFrame frame;
	private static CubesCanvas cubes;
	//private static int counter; // debugging

	int sizeX = 0, sizeY = 0, sizeZ = 0;
	int extraColumns = 0;
	int extraRows = 0;
	int extraStacks = 0;
	int layers;
	int[] tmpMatrix = null;
	int[] targetM = null;
	boolean[] locked = null;
	int[] locList = null;

	static final int MAX_PLAN = 1000;
	static final int QUEUE_SIZE = 1000;
	static final int COORD = 6;

	static final int[] otherDir = {
		CubesCanvas.BOTTOM, CubesCanvas.LEFT,
		CubesCanvas.TOP, CubesCanvas.RIGHT,
		CubesCanvas.OUTWARDS, CubesCanvas.INWARDS
	};

	CubesSolve(JFrame frame, CubesCanvas cubes) {
		CubesSolve.frame = frame;
		CubesSolve.cubes = cubes;
		//this.counter = 0;
	}

	/* puzzleSize MUST be a multiple of 2; */
	static int puzzleSize() {
		return ((java.lang.Math.max(java.lang.Math.max(
			java.lang.Math.min(cubes.blocks.x, cubes.blocks.y),
			java.lang.Math.min(cubes.blocks.y, cubes.blocks.z)),
			java.lang.Math.min(cubes.blocks.z, cubes.blocks.x))
			>>1)<<1);
	}

	/**
	 * layer info macros	- (innermost 4 tiles are layer zero,
	 *			ordinal goes up as you move out)
	 * layerHeight		- returns number of (rows down),
	 *			(columns across) the layer starts;
	 * layerWidth		- number of blocks wide the layer is;
	 */

	int layerHeight(int l) {
		return layers - 1 - l;
	}

	int layerWidth(int l) {
		return puzzleSize() - 2 * layerHeight(l);
	}

	/** macros for finding the corners of each layer */

	int UL_Z(int l) {
		return layerHeight(l) * (puzzleSize() + 1)
			+ extraRows * cubes.blocks.x
			+ extraColumns * (layers - l);
	}

	int UR_Z(int l) {
		return layerHeight(l) * (puzzleSize() + 1)
			+ layerWidth(l) - 1
			+ extraRows * cubes.blocks.x
			+ extraColumns * (layers - l);
	}

	int LL_Z(int l) {
		return (layerHeight(l) + layerWidth(l) - 1) * puzzleSize()
			+ layerHeight(l)
			+ extraRows * puzzleSize()
			+ extraColumns * (cubes.blocks.y + 1 + l - layers);
	}

	int LR_Z(int l) {
		return (layerHeight(l)
			+ layerWidth(l) - 1) * (puzzleSize() + 1)
			+ extraRows * puzzleSize()
			+ extraColumns * (cubes.blocks.y + 1 + l - layers);
	}

	int UL_Y(int l) {
		return layerHeight(l) * (puzzleSize() + 1)
			+ extraStacks * cubes.blocks.x
			+ extraColumns * (layers - l);
	}

	int UR_Y(int l) {
		return layerHeight(l) * (puzzleSize() + 1)
			+ layerWidth(l) - 1
			+ extraStacks * cubes.blocks.x
			+ extraColumns * (layers - l);
	}

	int LL_Y(int l) {
		return (layerHeight(l) + layerWidth(l) - 1) * puzzleSize()
			+ layerHeight(l)
			+ extraStacks * puzzleSize()
			+ extraColumns * (cubes.blocks.z + 1 + l - layers);
	}

	int LR_Y(int l) {
		return (layerHeight(l)
			+ layerWidth(l) - 1) * (puzzleSize() + 1)
			+ extraStacks * puzzleSize()
			+ extraColumns * (cubes.blocks.z + 1 + l - layers);
	}

	int UL_X(int l) {
		return layerHeight(l) * (puzzleSize() + 1)
			+ extraStacks * cubes.blocks.y
			+ extraRows * (layers - l);
	}

	int UR_X(int l) {
		return layerHeight(l) * (puzzleSize() + 1)
			+ layerWidth(l) - 1
			+ extraStacks * cubes.blocks.y
			+ extraRows * (layers - l);
	}

	int LL_X(int l) {
		return (layerHeight(l) + layerWidth(l) - 1) * puzzleSize()
			+ layerHeight(l)
			+ extraStacks * puzzleSize()
			+ extraRows * (cubes.blocks.z + 1 + l - layers);
	}

	int LR_X(int l) {
		return (layerHeight(l)
			+ layerWidth(l) - 1) * (puzzleSize() + 1)
			+ extraStacks * puzzleSize()
			+ extraRows * (cubes.blocks.z + 1 + l - layers);
	}

	/** get the x and y and z coordinates of a location in the matrix */

	static int toColumn(int loc) {
		return (loc % cubes.sizeRect) % cubes.blocks.x;
	}

	static int toRow(int loc) {
		return (loc % cubes.sizeRect) / cubes.blocks.x;
	}

	static int toStack(int loc) {
		return (loc / cubes.sizeRect);
	}

	static int toPosition(int x, int y, int z) {
		return z * cubes.sizeRect + y * cubes.blocks.x + x;
	}

	static int spaceX() {
		return toColumn(cubes.spacePosition);
	}

	static int spaceY() {
		return toRow(cubes.spacePosition);
	}

	static int spaceZ() {
		return toStack(cubes.spacePosition);
	}

	static int nextLeft(int loc) {
		return loc - 1;
	}

	static int nextRight(int loc) {
		return loc + 1;
	}

	static int nextUp(int loc) {
		return loc - cubes.blocks.x;
	}

	static int nextDown(int loc) {
		return loc + cubes.blocks.x;
	}

	static int nextIn(int loc) {
		return loc - cubes.sizeRect;
	}

	static int nextOut(int loc) {
		return loc + cubes.sizeRect;
	}

	static int findPiece(int piece) {
		int i;

		for (i = 0; i < cubes.sizeBlock; i++)
			if (cubes.blockOfPosition[i] == piece) {
				/*System.out.println("position=" + i
					+ ", piece=" + piece + " at "
					+ cubes.blockOfPosition[i]);*/
				return i;
			}

		System.err.print("Piece " + piece + " not found, exiting.");
		return -1;
	}

	static void logMoveSpace(int firstX, int firstY, int firstZ,
			int lastX, int lastY, int lastZ, int dir) {
		int count = java.lang.Math.abs(firstX - lastX) +
			java.lang.Math.abs(firstY - lastY) +
			java.lang.Math.abs(firstZ - lastZ) + 1;

		if (dir < 4)
			cubes.animateSlide(count, dir, CubesCanvas.DOUBLE, true);
		else if (dir == 4)
			for (int i = 1; i <= count; i++)
				cubes.discreteMoves(cubes.spacePosition + cubes.sizeRect, dir);
		else
			for (int i = 1; i <= count; i++)
				cubes.discreteMoves(cubes.spacePosition - cubes.sizeRect, dir);
	}

	void moveSpace(int direction, int distance) {
		int i, step, count;
		int firstX, firstY, firstZ;
		int lastX, lastY, lastZ, shiftDir;
		int tempSpaceX = spaceX();
		int tempSpaceY = spaceY();
		int tempSpaceZ = spaceZ();
		int dir = direction, dist = distance;

		if (!getFlag()) {
			return;
		}
		if (dist == 0)
			return;

		if (dir == CubesCanvas.LEFT) {
			dir = CubesCanvas.RIGHT;
			dist = -dist;
		} else if (dir == CubesCanvas.TOP) {
			dir = CubesCanvas.BOTTOM;
			dist = -dist;
		} else if (dir == CubesCanvas.INWARDS) {
			dir = CubesCanvas.OUTWARDS;
			dist = -dist;
		}
		firstX = spaceX();
		firstY = spaceY();
		firstZ = spaceZ();
		step = 1;
		count = dist;
		if (dist < 0) {
			step = -1;
			count = -count;
		}

		/*
		 * firstX,Y,Z are the location of the
		 * first piece to be shifted
		 */
		if (dir == CubesCanvas.RIGHT)
			firstX += step;
		else if (dir == CubesCanvas.BOTTOM)
			firstY += step;
		else /* if (dir == CubesCanvas.OUTWARDS) */
			firstZ += step;

		/* shiftDir is the direction the pieces need to be shifted */
		if (dist < 0) {
			shiftDir = dir;
		} else {
			shiftDir = otherDir[dir];
		}
		for (i = 0; i < count; i++) {
			if (dir == CubesCanvas.RIGHT) {
				tempSpaceX += step;
			} else if (dir == CubesCanvas.BOTTOM) {
				tempSpaceY += step;
			} else { /* if (dir == CubesCanvas.OUTWARDS) */
				tempSpaceZ += step;
			}
		}
		lastX = tempSpaceX;
		lastY = tempSpaceY;
		lastZ = tempSpaceZ;

		/*
		 * the blocks firstX,Y,Z through lastX,Y,Z need to be shifted
		 * one block in the shiftDir direction;
		 */

		logMoveSpace(firstX, firstY, firstZ,
			lastX, lastY, lastZ, shiftDir);
	}

	static void printIntMatrix(int[] mat) {
		int i, j, k;

		System.out.println("intMatrix");
		for (k = 0; k < cubes.blocks.z; k++) {
			for (j = 0; j < cubes.blocks.y; j++) {
				for (i = 0; i < cubes.blocks.x; i++) {
					if (mat[toPosition(i, j, k)] < 10 &&
							mat[toPosition(i, j, k)] >= 0)
						System.out.print(" ");
					if (mat[toPosition(i, j, k)] < 100)
						System.out.print(" ");
					System.out.print(" " + mat[toPosition(i, j, k)]);
				}
				System.out.println();
			}
			System.out.println();
		}
		System.out.println();
	}

	static void printBooleanMatrix(boolean[] mat) {
		int i, j, k;

		System.out.println("booleanMatrix");
		for (k = 0; k < cubes.blocks.z; k++) {
			for (j = 0; j < cubes.blocks.y; j++) {
				for (i = 0; i < cubes.blocks.x; i++) {
					System.out.print(" ");
					System.out.print((mat[toPosition(i, j, k)]) ?
						1 : 0);
				}
				System.out.println();
			}
			System.out.println();
		}
		System.out.println();
	}

	static int Dist(int loc1, int loc2) {
		return (java.lang.Math.abs(toColumn(loc1) - toColumn(loc2)) +
			java.lang.Math.abs(toRow(loc1) - toRow(loc2)) +
			java.lang.Math.abs(toStack(loc1) - toStack(loc2)));
	}

	void planMove(int startLoc, int endLoc, int[] path) {
		int i, nextLoc = 0, nextDist, chosen, moveNum;
		boolean foundPath = false, debug = false;
		int locX, locY, locZ;
		int[] locQueue = new int[QUEUE_SIZE];
		int[] locDist = new int[QUEUE_SIZE];
		boolean[] locQueueUsed = new boolean[QUEUE_SIZE];
		int queueHead = 0, queueTail = 0;
		int[] candidate = new int[COORD];

		for (i = 0; i < cubes.sizeBlock; i++) {
			tmpMatrix[i] = (locked[i]) ? -1 : 0;
			locList[i] = -1;
		}

		for (i = 0; i < QUEUE_SIZE; i++)
			locQueueUsed[i] = false;

		locQueue[0] = startLoc;
		locDist[0] = Dist(endLoc, startLoc);
		tmpMatrix[startLoc] = 1;
		queueTail++;

		/*
		 * if the selected element has a distance of zero,
		 * we have found it; (This really is not a queue,
		 * but rather a range of elements to be searched
		 * for an element of the desired properties;
		 */

		/*
		 * as we search for a path,
		 * LINK array is used to indicate the direction from which
		 * we moved into a location;
		 * TMP_MATRIX array is used to keep track of the move number;
		 */

		while (queueHead < queueTail && !foundPath) {
			/*
			 * find the entry that
			 * (1) has the smallest distance and
			 * (2) has the smallest move number;
			 */

			nextLoc = locQueue[queueHead];
			nextDist = locDist[queueHead];
			chosen = queueHead;

			for (i = queueHead + 1; i < queueTail; i++)
				if (!locQueueUsed[i] &&
						((locDist[i] < nextDist) ||
						((locDist[i] == nextDist) &&
						(tmpMatrix[locQueue[i]] <
						tmpMatrix[nextLoc])))) {
					nextLoc = locQueue[i];
					nextDist = locDist[i];
					chosen = i;
				}
			if (nextDist == 0) {
				foundPath = true;
				break;
			}
			locQueueUsed[chosen] = true;

			/* permute the chosen element */
			candidate[CubesCanvas.TOP] = nextUp(nextLoc);
			candidate[CubesCanvas.LEFT] = nextLeft(nextLoc);
			candidate[CubesCanvas.BOTTOM] = nextDown(nextLoc);
			candidate[CubesCanvas.RIGHT] = nextRight(nextLoc);
			candidate[CubesCanvas.INWARDS] = nextIn(nextLoc);
			candidate[CubesCanvas.OUTWARDS] = nextOut(nextLoc);
			locX = toColumn(nextLoc);
			locY = toRow(nextLoc);
			locZ = toStack(nextLoc);
			if (locX == 0)
				candidate[CubesCanvas.LEFT] = -1;
			if (locX == cubes.blocks.x - 1)
				candidate[CubesCanvas.RIGHT] = -1;
			if (locY == 0)
				candidate[CubesCanvas.TOP] = -1;
			if (locY == cubes.blocks.y - 1)
				candidate[CubesCanvas.BOTTOM] = -1;
			if (locZ == 0)
				candidate[CubesCanvas.INWARDS] = -1;
			if (locZ == cubes.blocks.z - 1)
				candidate[CubesCanvas.OUTWARDS] = -1;
			moveNum = tmpMatrix[nextLoc] + 1;
			for (i = 0; i < COORD; i++)
				if (candidate[i] != -1 &&
						tmpMatrix[candidate[i]] == 0) {
					tmpMatrix[candidate[i]] = moveNum;
					/*
					 * the next line works because the
					 * candidate index is same as the
					 * direction moved to reach the
					 * candidate;
					 */
					locList[candidate[i]] = i;
					locQueue[queueTail] = candidate[i];
					locDist[queueTail] =
						Dist(endLoc, candidate[i]);
					queueTail++;
					if (queueTail == QUEUE_SIZE) {
						System.err.println(
				"Queue size is exceeded, exiting.");
						return;
					}
				}

			/* delete used items from the front of the queue */
			while (locQueueUsed[queueHead] &&
					queueHead < queueTail)
				queueHead++;
		}
		if (debug) {
			printIntMatrix(cubes.blockOfPosition);
			printIntMatrix(targetM);
			printBooleanMatrix(locked);
			printIntMatrix(locList);
			if (!foundPath) {
				System.err.print("Could not find a way to move ");
				System.err.print(startLoc);
				System.err.print(" (" + toColumn(startLoc));
				System.err.print("," + toRow(startLoc));
				System.err.print("," + toStack(startLoc));
				System.err.print(") to " + endLoc);
				System.err.print(" (" + toColumn(endLoc));
				System.err.print("," + toRow(endLoc));
				System.err.print("," + toStack(endLoc));
				System.err.println("), exiting.");
				return;
			}
		} else {
			if (!foundPath) {
				return;
			}
		}

		/*
		 * copy the path we found into the path array;
		 * element 0 will contain the number of moves in the path;
		 * by the time we get there,
		 * nextLoc is in the final location
		 */

		path[0] = tmpMatrix[nextLoc] - 1;
		for (i = path[0]; i > 0; i--) {
			path[i] = locList[nextLoc];
			switch(locList[nextLoc]) {
			case CubesCanvas.LEFT:
				nextLoc = nextRight(nextLoc);
				break;
			case CubesCanvas.RIGHT:
				nextLoc = nextLeft(nextLoc);
				break;
			case CubesCanvas.TOP:
				nextLoc = nextDown(nextLoc);
				break;
			case CubesCanvas.BOTTOM:
				nextLoc = nextUp(nextLoc);
				break;
			case CubesCanvas.INWARDS:
				nextLoc = nextOut(nextLoc);
				break;
			case CubesCanvas.OUTWARDS:
				nextLoc = nextIn(nextLoc);
				break;
			default:
			}
		}
	}

	void myMoveSpaceTo(int loc) {
		int i, currentDir, dist = 0;
		int[] plan = new int[MAX_PLAN];

		planMove(toPosition(spaceX(), spaceY(), spaceZ()), loc, plan);
		currentDir = plan[1];
		for (i = 1; i <= plan[0]; i++) {
			if (plan[i] == currentDir) {
				dist++;
			} else if (plan[i] == otherDir[currentDir]) {
				dist--;
			} else {
				moveSpace(currentDir, dist);
				currentDir = plan[i];
				dist = 1;
			}
		}
		moveSpace(currentDir, dist);
	}

	void movePiece(int location, int targetMove) {
		int i;
		int[] plan = new int[MAX_PLAN];
		int loc = location;

		planMove(loc, targetMove, plan);
		for (i = 1; i <= plan[0]; i++)
			switch(plan[i]) {
			case CubesCanvas.LEFT:
				locked[loc] = true;
				myMoveSpaceTo(nextLeft(loc));
				locked[loc] = false;
				myMoveSpaceTo(loc);
				loc = nextLeft(loc);
				break;
			case CubesCanvas.RIGHT:
				locked[loc] = true;
				myMoveSpaceTo(nextRight(loc));
				locked[loc] = false;
				myMoveSpaceTo(loc);
				loc = nextRight(loc);
				break;
			case CubesCanvas.TOP:
				locked[loc] = true;
				myMoveSpaceTo(nextUp(loc));
				locked[loc] = false;
				myMoveSpaceTo(loc);
				loc = nextUp(loc);
				break;
			case CubesCanvas.BOTTOM:
				locked[loc] = true;
				myMoveSpaceTo(nextDown(loc));
				locked[loc] = false;
				myMoveSpaceTo(loc);
				loc = nextDown(loc);
				break;
			case CubesCanvas.INWARDS:
				locked[loc] = true;
				myMoveSpaceTo(nextIn(loc));
				locked[loc] = false;
				myMoveSpaceTo(loc);
				loc = nextIn(loc);
				break;
			case CubesCanvas.OUTWARDS:
				locked[loc] = true;
				myMoveSpaceTo(nextOut(loc));
				locked[loc] = false;
				myMoveSpaceTo(loc);
				loc = nextOut(loc);
				break;
			default:
			}
	}

	void myInitialize() {
		/*
		 * Initialize the position and
		 * the targetM matrices;
		 */

		int i;
		int spX, spY, spZ;

		sizeX = cubes.blocks.x;
		sizeY = cubes.blocks.y;
		sizeZ = cubes.blocks.z;
		layers = puzzleSize() >> 1;

		extraColumns = cubes.blocks.x - puzzleSize();
		extraRows = cubes.blocks.y - puzzleSize();
		extraStacks = cubes.blocks.z - puzzleSize();

		tmpMatrix = new int[cubes.sizeBlock];
		targetM = new int[cubes.sizeBlock];
		locked = new boolean[cubes.sizeBlock];
		locList = new int[cubes.sizeBlock];

		for (i = 0; i < cubes.sizeBlock; i++)
			locked[i] = false;

		if (tmpMatrix == null || targetM == null ||
				locked == null || locList == null) {
			System.err.print("Not enough memory, exiting.");
		}

		for (i = 0; i < cubes.sizeBlock - 1; i++) {
			targetM[i] = i + 1;
		}
		targetM[cubes.sizeBlock - 1] = 0;

		/*
		 * Move the space into the LR_* corner of the
		 * innermost layer;
		 * For each of the outer layers, move the space
		 * left one and up one;
		 */

		spX = cubes.blocks.x - 1;
		spY = cubes.blocks.y - 1;
		spZ = cubes.blocks.z - 1;

		/* Make solution with space at center bottom right */
		for (i = 0; i < layers - 1; i++) {
			if (spX > 0) {
				/* move the space left one; */
				targetM[toPosition(spX, spY, spZ)] =
					targetM[toPosition((spX - 1), spY, spZ)];
				targetM[toPosition((spX - 1), spY, spZ)] = 0;
				spX -= 1;
			}
			if (spY > 0) {
				/* move the space up one; */
				targetM[toPosition(spX, spY, spZ)] =
					targetM[toPosition(spX, (spY - 1), spZ)];
				targetM[toPosition(spX, (spY - 1), spZ)] = 0;
				spY -= 1;
			}
			if (spZ > 0) {
				/* move the space in one; */
				targetM[toPosition(spX, spY, spZ)] =
					targetM[toPosition(spX, spY, (spZ - 1))];
				targetM[toPosition(spX, spY, (spZ - 1))] = 0;
				spZ -= 1;
			}
		}
	}

	/*
	 * To solve this puzzle, work from the outside in;
	 * For each successive ring working your way in,
	 *
	 * (1) put the corners in place;
	 * (2) finish off the rest of the boundaries;
	 * (3) do the next layer in;
	 */

	void solveLayerZ0() {
		movePiece(findPiece(targetM[UL_Z(0)]), UL_Z(0));
		myMoveSpaceTo(LR_Z(0));
	}

	void solveLayerY0() {
		movePiece(findPiece(targetM[UL_Y(0)]), UL_Y(0));
		myMoveSpaceTo(LR_Y(0));
	}

	void solveLayerX0() {
		movePiece(findPiece(targetM[UL_X(0)]), UL_X(0));
		myMoveSpaceTo(LR_X(0));
	}

	void doLastTwoOnEdge(int ntLast, int last, int tmp, int emergency) {
		int lastPiece, ntLastPiece;

		lastPiece = targetM[last];
		ntLastPiece = targetM[ntLast];

		movePiece(findPiece(ntLastPiece), last);
		locked[last] = true;

		/*
		 * if the last piece is stuck where the next to the last
		 * piece should go, do some magic to fix things up;
		 */
		if (findPiece(0) == ntLast)
			myMoveSpaceTo(tmp);

		if (findPiece(lastPiece) == ntLast) {
			/* a rescue is necessary */
			locked[last] = false;
			movePiece(findPiece(ntLastPiece), ntLast);
			locked[ntLast] = true;
			movePiece(findPiece(lastPiece), emergency);
			locked[emergency] = true;
			locked[ntLast] = false;
			movePiece(findPiece(ntLastPiece), last);
			locked[emergency] = false;
			locked[last] = true;
		}

		movePiece(findPiece(lastPiece), tmp);
		locked[tmp] = true;
		myMoveSpaceTo(ntLast);
		locked[tmp] = false;
		locked[last] = false;
		myMoveSpaceTo(last);
		myMoveSpaceTo(tmp);
		locked[ntLast] = true;
		locked[last] = true;
	}

	void solveLayerZ(int layer) {
		int i, tmp, last, ntLast, emergency;
		int ul, ur, ll, lr;

		if (layer == 0) {
			solveLayerZ0();
		} else {
			/* find and put each of the corners into place */
			ul = UL_Z(layer);
			ur = UR_Z(layer);
			ll = LL_Z(layer);
			lr = LR_Z(layer);

			movePiece(findPiece(targetM[ul]), ul);
			locked[ul] = true;
			movePiece(findPiece(targetM[ur]), ur);
			locked[ur] = true;
			movePiece(findPiece(targetM[ll]), ll);
			locked[ll] = true;
			movePiece(findPiece(targetM[lr]), lr);
			locked[lr] = true;

			/*
			 * Strategy for doing the pieces between the corners:
			 * (1) put all but the last two edge pieces in place;
			 * (2) put the next to the last piece next to the
			 *	corner;
			 * (3) put the last piece one move in from its final
			 *	position;
			 * (4) move the space to the final position of the
			 *	next to the last piece;
			 * (5) slide the next to the last piece over and the
			 *	last piece into the edge where it goes.
			 */

			/* top edge */
			for (i = ul + 1; i < ur - 2; i++) {
				movePiece(findPiece(targetM[i]), i);
				locked[i] = true;
			}
			ntLast = i;
			last = i + 1;
			tmp = UR_Z(layer - 1);
			emergency = nextDown(tmp);
			doLastTwoOnEdge(ntLast, last, tmp, emergency);

			/* bottom edge */
			for (i = ll + 1; i < lr - 2; i++) {
				movePiece(findPiece(targetM[i]), i);
				locked[i] = true;
			}

			ntLast = i;
			last = i + 1;
			tmp = LR_Z(layer - 1);
			emergency = nextUp(tmp);
			doLastTwoOnEdge(ntLast, last, tmp, emergency);

			/* left side */
			for (i = ul + cubes.blocks.x;
					i < ll - 2 * cubes.blocks.x;
					i += cubes.blocks.x) {
				movePiece(findPiece(targetM[i]), i);
				locked[i] = true;
			}

			ntLast = i;
			last = i + cubes.blocks.x;
			tmp = LL_Z(layer - 1);
			emergency = nextRight(tmp);
			doLastTwoOnEdge(ntLast, last, tmp, emergency);

			/* right side */
			for (i = ur + cubes.blocks.x;
					i < lr - 2 * cubes.blocks.x;
					i += cubes.blocks.x) {
				movePiece(findPiece(targetM[i]), i);
				locked[i] = true;
			}

			ntLast = i;
			last = i + cubes.blocks.x;
			tmp = LR_Z(layer - 1);
			emergency = nextLeft(tmp);
			doLastTwoOnEdge(ntLast, last, tmp, emergency);
		}
	}

	void solveLayerY(int layer) {
		int i, tmp, last, ntLast, emergency;
		int ul, ur, ll, lr;

		if (layer == 0) {
			solveLayerY0();
		} else {
			/* find and put each of the corners into place */
			ul = UL_Y(layer);
			ur = UR_Y(layer);
			ll = LL_Y(layer);
			lr = LR_Y(layer);

			movePiece(findPiece(targetM[ul]), ul);
			locked[ul] = true;
			movePiece(findPiece(targetM[ur]), ur);
			locked[ur] = true;
			movePiece(findPiece(targetM[ll]), ll);
			locked[ll] = true;
			movePiece(findPiece(targetM[lr]), lr);
			locked[lr] = true;

			/*
			 * Strategy for doing the pieces between the corners:
			 * (1) put all but the last two edge pieces in place;
			 * (2) put the next to the last piece next to the
			 *	corner;
			 * (3) put the last piece one move in from its final
			 *	position;
			 * (4) move the space to the final position of the
			 *	next to the last piece;
			 * (5) slide the next to the last piece over and the
			 *	last piece into the edge where it goes.
			 */

			/* inner edge */
			for (i = ul + 1; i < ur - 2; i++) {
				movePiece(findPiece(targetM[i]), i);
				locked[i] = true;
			}

			ntLast = i;
			last = i + 1;
			tmp = UR_Y(layer - 1);
			emergency = nextOut(tmp);
			doLastTwoOnEdge(ntLast, last, tmp, emergency);

			/* outer edge */
			for (i = ll + 1; i < lr - 2; i++) {
				movePiece(findPiece(targetM[i]), i);
				locked[i] = true;
			}

			ntLast = i;
			last = i + 1;
			tmp = LR_Y(layer - 1);
			emergency = nextIn(tmp);
			doLastTwoOnEdge(ntLast, last, tmp, emergency);

			/* left side */
			for (i = ul + cubes.blocks.x;
					i < ll - 2 * cubes.blocks.x;
					i += cubes.blocks.x) {
				movePiece(findPiece(targetM[i]), i);
				locked[i] = true;
			}

			ntLast = i;
			last = i + cubes.blocks.x;
			tmp = LL_Y(layer - 1);
			emergency = nextRight(tmp);
			doLastTwoOnEdge(ntLast, last, tmp, emergency);

			/* right side */
			for (i = ur + cubes.blocks.x;
					i < lr - 2 * cubes.blocks.x;
					i += cubes.blocks.x) {
				movePiece(findPiece(targetM[i]), i);
				locked[i] = true;
			}

			ntLast = i;
			last = i + cubes.blocks.x;
			tmp = LR_Y(layer - 1);
			emergency = nextLeft(tmp);
			doLastTwoOnEdge(ntLast, last, tmp, emergency);
		}
	}

	void solveLayerX(int layer) {
		int i, tmp, last, ntLast, emergency;
		int ul, ur, ll, lr;

		if (layer == 0) {
			solveLayerX0();
		} else {
			/* find and put each of the corners into place */
			ul = UL_X(layer);
			ur = UR_X(layer);
			ll = LL_X(layer);
			lr = LR_X(layer);

			movePiece(findPiece(targetM[ul]), ul);
			locked[ul] = true;
			movePiece(findPiece(targetM[ur]), ur);
			locked[ur] = true;
			movePiece(findPiece(targetM[ll]), ll);
			locked[ll] = true;
			movePiece(findPiece(targetM[lr]), lr);
			locked[lr] = true;

			/*
			 * Strategy for doing the pieces between the corners:
			 * (1) put all but the last two edge pieces in place;
			 * (2) put the next to the last piece next to the
			 *	corner;
			 * (3) put the last piece one move in from its final
			 *	position;
			 * (4) move the space to the final position of the
			 *	next to the last piece;
			 * (5) slide the next to the last piece over and the
			 *	last piece into the edge where it goes.
			 */

			/* inner edge */
			for (i = ul + 1; i < ur - 2; i++) {
				movePiece(findPiece(targetM[i]), i);
				locked[i] = true;
			}

			ntLast = i;
			last = i + 1;
			tmp = UR_X(layer - 1);
			emergency = nextOut(tmp);
			doLastTwoOnEdge(ntLast, last, tmp, emergency);

			/* outer edge */
			for (i = ll + 1; i < lr - 2; i++) {
				movePiece(findPiece(targetM[i]), i);
				locked[i] = true;
			}

			ntLast = i;
			last = i + 1;
			tmp = LR_X(layer - 1);
			emergency = nextIn(tmp);
			doLastTwoOnEdge(ntLast, last, tmp, emergency);

			/* top edge */
			for (i = ul + cubes.blocks.y;
					i < ll - 2 * cubes.blocks.y;
					i += cubes.blocks.y) {
				movePiece(findPiece(targetM[i]), i);
				locked[i] = true;
			}

			ntLast = i;
			last = i + cubes.blocks.y;
			tmp = LL_X(layer - 1);
			emergency = nextDown(tmp);
			doLastTwoOnEdge(ntLast, last, tmp, emergency);

			/* bottom edge */
			for (i = ur + cubes.blocks.y;
					i < lr - 2 * cubes.blocks.y;
					i += cubes.blocks.y) {
				movePiece(findPiece(targetM[i]), i);
				locked[i] = true;
			}

			ntLast = i;
			last = i + cubes.blocks.y;
			tmp = LR_X(layer - 1);
			emergency = nextUp(tmp);
			doLastTwoOnEdge(ntLast, last, tmp, emergency);
		}
	}

	void solveRowZ(int row, int z) {
		int i, loc, last, ntLast, tmp, emergency;

		for (i = 0; i < cubes.blocks.x - 2; i++) {
			loc = toPosition(i, row, z);
			movePiece(findPiece(targetM[loc]), loc);
			locked[loc] = true;
		}
		ntLast = toPosition(cubes.blocks.x - 2, row, z);
		last = toPosition(cubes.blocks.x - 1, row, z);
		tmp = last + cubes.blocks.x;
		emergency = tmp + cubes.blocks.x;
		doLastTwoOnEdge(ntLast, last, tmp, emergency);
	}

	void solveColumnZ(int column, int z) {
		int i, loc, last, ntLast, tmp, emergency;

		for (i = 0; i < cubes.blocks.y - 2; i++) {
			loc = toPosition(column, i, z);
			movePiece(findPiece(targetM[loc]), loc);
			locked[loc] = true;
		}
		ntLast = toPosition(column, cubes.blocks.y - 2, z);
		last = toPosition(column, cubes.blocks.y - 1, z);
		tmp = last + 1;
		emergency = tmp + 1;
		doLastTwoOnEdge(ntLast, last, tmp, emergency);
	}

	void solveStackY(int stack, int y) {
		int i, loc, last, ntLast, tmp, emergency;

		for (i = 0; i < cubes.blocks.x - 2; i++) {
			loc = toPosition(i, y, stack);
			movePiece(findPiece(targetM[loc]), loc);
			locked[loc] = true;
		}
		ntLast = toPosition(cubes.blocks.x - 2, y, stack);
		last = toPosition(cubes.blocks.x - 1, y, stack);
		tmp = last + cubes.blocks.x;
		emergency = tmp + cubes.blocks.x;
		doLastTwoOnEdge(ntLast, last, tmp, emergency);
	}

	void solveColumnY(int column, int y) {
		int i, loc, last, ntLast, tmp, emergency;

		for (i = 0; i < cubes.blocks.z - 2; i++) {
			loc = toPosition(column, y, i);
			movePiece(findPiece(targetM[loc]), loc);
			locked[loc] = true;
		}
		ntLast = toPosition(column, y, cubes.blocks.z - 2);
		last = toPosition(column, y, cubes.blocks.z - 1);
		tmp = last + 1;
		emergency = tmp + 1;
		doLastTwoOnEdge(ntLast, last, tmp, emergency);
	}

	void solveStackX(int stack, int x) {
		int i, loc, last, ntLast, tmp, emergency;

		for (i = 0; i < cubes.blocks.y - 2; i++) {
			loc = toPosition(x, i, stack);
			movePiece(findPiece(targetM[loc]), loc);
			locked[loc] = true;
		}
		ntLast = toPosition(x, cubes.blocks.y - 2, stack);
		last = toPosition(x, cubes.blocks.y - 1, stack);
		tmp = last + cubes.blocks.y;
		emergency = tmp + cubes.blocks.y;
		doLastTwoOnEdge(ntLast, last, tmp, emergency);
	}

	void solveRowX(int row, int x) {
		int i, loc, last, ntLast, tmp, emergency;

		for (i = 0; i < cubes.blocks.z - 2; i++) {
			loc = toPosition(x, row, i);
			movePiece(findPiece(targetM[loc]), loc);
			locked[loc] = true;
		}
		ntLast = toPosition(x, row, cubes.blocks.z - 2);
		last = toPosition(x, row, cubes.blocks.z - 1);
		tmp = last + 1;
		emergency = tmp + 1;
		doLastTwoOnEdge(ntLast, last, tmp, emergency);
	}

	void solveSinglePlaneZ(int plane) {
		int i;

		/* solve the extra rows and columns */
		for (i = 0; i < extraRows; i++)
			solveRowZ(i, plane);
		for (i = 0; i < extraColumns; i++)
			solveColumnZ(i, plane);

		/* solve each layer */
		for (i = layers - 1; i >= 0; i--) {
			solveLayerZ(i);
		}
		/* move the space back out to the LR_Z corner; */
		/* i is the layer the space is moving into */
		for (i = 1; i < layers; i++) {
			moveSpace(CubesCanvas.BOTTOM, 1);
			moveSpace(CubesCanvas.RIGHT, 1);
		}
	}

	void solveSinglePlaneY(int plane) {
		int i;

		/* solve the extra stacks and rows */
		for (i = 0; i < extraStacks; i++)
			solveStackY(i, plane);
		for (i = 0; i < extraColumns; i++)
			solveColumnY(i, plane);

		/* solve each layer */
		for (i = layers - 1; i >= 0; i--) {
			solveLayerY(i);
		}
		/* move the space back out to the LR_Y corner; */
		/* i is the layer the space is moving into */
		for (i = 1; i < layers; i++) {
			moveSpace(CubesCanvas.OUTWARDS, 1);
			moveSpace(CubesCanvas.RIGHT, 1);
		}
	}

	void solveSinglePlaneX(int plane) {
		int i;

		/* solve the extra stacks and rows */
		for (i = 0; i < extraStacks; i++)
			solveStackX(i, plane);
		for (i = 0; i < extraRows; i++)
			solveRowX(i, plane);

		/* solve each layer */
		for (i = layers - 1; i >= 0; i--) {
			solveLayerX(i);
		}
		/* move the space back out to the LR_X corner; */
		/* i is the layer the space is moving into */
		for (i = 1; i < layers; i++) {
			moveSpace(CubesCanvas.OUTWARDS, 1);
			moveSpace(CubesCanvas.BOTTOM, 1);
		}
	}

	/* This procedure coordinates the solution process. */
	public void solveSomeBlocks() {
		int i;

		if (sizeX != cubes.blocks.x || sizeY != cubes.blocks.y ||
				sizeZ != cubes.blocks.z) {
			myInitialize();
		}
		((CubesFrame) frame).callback(
			CubesInterface.ACTION_RESET);
		/*
		 * determine the position we want to be in when
		 * we are done; This position will have the space in
		 * the center; Then, we'll move the space back to
		 * the outside.
		 */
		if (!cubes.checkSolved()) {
			 if (sizeY == 1 && sizeZ == 1)
				moveSpace(CubesCanvas.RIGHT,
					sizeX - spaceX() - 1);
			else if (sizeX == 1 && sizeZ == 1)
				moveSpace(CubesCanvas.BOTTOM,
					sizeY - spaceY() - 1);
			else if (sizeX == 1 && sizeY == 1)
				moveSpace(CubesCanvas.OUTWARDS,
					sizeZ - spaceZ() - 1);
			else if (sizeZ == 1) {
				solveSinglePlaneZ(0);
			} else if (sizeY == 1) {
				solveSinglePlaneY(0);
			} else if (sizeX == 1) {
				solveSinglePlaneX(0);
			} else { /* This fails but for 3x3x3 solves a few tiles */
				for (i = 0; i < sizeZ; i++)
					solveSinglePlaneZ(i);
			}
		}
		((CubesFrame) frame).callback(
			CubesInterface.ACTION_COMPUTED);
	}

	private class SeparateSubTask extends Thread {
		private boolean runFlag = false;
		SeparateSubTask() {
		}
		public void init() {
			start();
		}
		public void setFlag(boolean flag) {
			runFlag = flag;
		}
		public boolean getFlag() {
			return runFlag;
		}
		@Override
		public void run() {
			while (true) {
				try {
					sleep(100);
				} catch (InterruptedException e) {
					System.err.println("Interrupted");
				}
				if (runFlag) {
					solveSomeBlocks();
					runFlag = false;
				}
			}
		}
	}
	private SeparateSubTask sp = null;
	public void setFlag(boolean flag) {
		if (sp != null)
			sp.setFlag(flag);
	}
	public boolean getFlag() {
		if (sp == null)
			return false;
		return sp.getFlag();
	}
	public void init() {
		if (sp == null) {
			sp = new SeparateSubTask();
			sp.init();
		}
	}
}
